/* usadelndsoc
 *
 * Copyright © 2022 Pauli Virtanen
 *    @author Pauli Virtanen <pauli.t.virtanen@jyu.fi>
 *
 * SPDX-License-Identifier: AGPL-3.0-or-later
 */
#include <unordered_map>
#include <functional>

#include <pybind11/pybind11.h>
#include <pybind11/complex.h>

#include "common.hpp"
#include "array.hpp"
#include "action.hpp"

#include "pythonutil.hpp"

using namespace array;
using namespace array::python;

#define CLASS_Action PY_VISIBILITY Action
class CLASS_Action
{
public:
    typedef PyArrayView<const Complex,Shape<Dynamic,Dynamic,4,4,4> > UType;
    typedef PyArrayView<const Complex,Shape<Dynamic,Dynamic,4,4> > OmegaType;
    typedef PyArrayView<const Complex,Shape<Dynamic,Dynamic,2,2,2> > QType;
    typedef PyArrayView<const Mask,Shape<Dynamic,Dynamic> > MaskType;

    struct pair_hash {
        size_t operator () (const std::pair<size_t, size_t>& p) const
            {
                size_t h1 = std::hash<size_t>{}(p.first);
                size_t h2 = std::hash<size_t>{}(p.second);
                size_t golden_ratio;
                if constexpr (sizeof(size_t) < 8) {
                    golden_ratio = 0x9e3779b9;
                } else {
                    golden_ratio = 0x9e3779b97f4a7c15;
                }
                h1 ^= h2 + golden_ratio + (h1 << 6) + (h1 >> 2);
                return h1;
            }
    };

private:
    UType U_;
    OmegaType Omega_;
    MaskType mask_;
    size_t nx_, ny_;
    double Lx_, Ly_;
    Complex alpha_;
    Complex D_;
    Complex omega_;

    typedef Vector<size_t> SizeVector;
    typedef Vector<Complex> ComplexVector;
    typedef std::vector<bool> BoolVector;
    typedef CppAD::sparse_rc<SizeVector> Sparsity;

    CppAD::ADFun<Complex> f_;
    CppAD::sparse_hes_work hes_work_;
    Sparsity hes_pattern_;
    CppAD::sparse_rcv<SizeVector, ComplexVector> hes_subset_;
    bool computed_ = false;
    bool hes_computed_ = false;

    std::unordered_map<std::pair<size_t,size_t>, bool, pair_hash> hes_pattern_map_;

    pybind11::object py_sparse_matrix_type_;

    Vector<Complex> ravel(const QType& Q)
        {
            if (Q.dim(0) != nx_ || Q.dim(1) != ny_)
                throw std::domain_error("input Q has mismatching shape");

            Complex *data = const_cast<Complex *>(Q.data());
            size_t size = Q.size();
            return std::move(Vector<Complex>(data, size));
        }

    void set_dynamic()
        {
            std::vector<Complex> dynamic(1);
            dynamic[0] = omega_;
            f_.new_dynamic(dynamic);
        }

public:
    Action(double Lx, double Ly, Complex alpha, Complex D, MaskType&& mask, UType&& U, OmegaType&& Omega)
        : U_(std::move(U)), Omega_(std::move(Omega)), mask_(std::move(mask)), nx_(U_.dim(0)), ny_(U_.dim(1)), Lx_(Lx), Ly_(Ly), alpha_(alpha), D_(D), omega_(0),
          hes_pattern_map_()
        {
            if (Omega.dim(0) != nx_ || Omega.dim(1) != ny_)
                throw std::domain_error("U and Omega have incompatible shape");
            if (mask.dim(0) != nx_ || mask.dim(1) != ny_)
                throw std::domain_error("U and mask have incompatible shape");

            pybind11::object sparse_module = pybind11::module_::import("scipy.sparse");
            py_sparse_matrix_type_ = sparse_module.attr("coo_matrix");
        }

    void set_omega(const Complex& omega) { omega_ = omega; }
    const Complex& get_omega() const { return omega_; }

    Complex eval(const QType& Q)
        {
            if (Q.dim(0) != nx_ || Q.dim(1) != ny_)
                throw std::out_of_range("Q array wrong size");
            return S(mask_, Q, omega_, U_, Omega_, Lx_, Ly_, alpha_, D_);
        }

    CppAD::ADFun<Complex> eval_ad(const QType& Qval)
        {
            if (Qval.dim(0) != nx_ || Qval.dim(1) != ny_)
                throw std::out_of_range("Q array wrong size");

            Array<ADComplex, Shape<Dynamic,Dynamic,2,2,2> >
                Q({nx_, ny_, 2, 2, 2});

            auto a = Qval.reshape<Dynamic>({Qval.size()});
            auto b = Q.reshape<Dynamic>({Q.size()});
            for (size_t i = 0; i < a.dim(0); ++i)
                b(i) = a(i);

            std::vector<ADComplex> dynamic(1);
            dynamic[0] = omega_;

            CppAD::Independent(Q.storage(), dynamic);

            Array<ADComplex, Shape<1> > res;
            res(0u) = S(mask_, Q, dynamic[0], U_, Omega_, Lx_, Ly_, alpha_, D_);

            return {Q.storage(), res.storage()};
        }

    void compute(const QType& Q)
        {
            f_ = eval_ad(Q);
            f_.optimize();

            /*
             * Hessian sparsity pattern
             */

            hes_pattern_.resize(0, 0, 0);
            hes_work_.clear();
            hes_computed_ = false;

            /* Right-multiplier is identity matrix */
            size_t n      = Q.size();
            size_t nr     = n;
            size_t nc     = n;
            size_t nnz_in = n;
            Sparsity pattern_in(nr, nc, nnz_in);
            for(size_t k = 0; k < nnz_in; k++)
            {
                size_t r = k;
                size_t c = k;
                pattern_in.set(k, r, c);
            }

            /* Hessian sparsity: */
            BoolVector s = {true};
            f_.for_jac_sparsity(pattern_in, false, false, false, hes_pattern_);
            f_.rev_hes_sparsity(s, false, false, hes_pattern_);
            f_.size_forward_set(0);

            hes_subset_ = hes_pattern_;

            /* Update map */
            hes_pattern_map_.clear();
            auto r = hes_pattern_.row();
            auto c = hes_pattern_.col();
            for (size_t k = 0; k < hes_pattern_.nnz(); ++k)
                hes_pattern_map_.emplace(std::make_pair(std::pair(r[k], c[k]), true));

            computed_ = true;
        }

    auto grad(const QType& Q)
        {
            Vector<Complex> x = ravel(Q);

            if (!computed_)
                compute(Q);
            else
                set_dynamic();

            auto jac = f_.Jacobian(x);
            return Array<Complex, Shape<Dynamic>, Vector<Complex> >(std::move(jac), {jac.size()});
        }

    auto hess_mul(const QType& Q, const QType& dQ)
        {
            Vector<Complex> x = ravel(Q);
            Vector<Complex> dx = ravel(dQ);
            size_t n = x.size();

            if (!computed_)
                compute(Q);
            else
                set_dynamic();

            Vector<Complex> w(1);
            w[0] = 1;

            f_.Forward(0, x);
            f_.Forward(1, dx);
            auto ddw = f_.Reverse(2, w);

            if (ddw.size() != 2*n)
                throw std::domain_error("wrong output size");

            Vector<Complex> v(n);

            for(size_t j = 0; j < n; j++)
                v[j] = ddw[2*j + 1];

            return Array<Complex, Shape<Dynamic>, Vector<Complex> >(std::move(v), {v.size()});
        }

    auto hess(const QType& Q)
        {
            Vector<Complex> x = ravel(Q);

            if (!computed_)
                compute(Q);
            else
                set_dynamic();

            /* Hessian */
            ComplexVector w = {1.0};
            std::string coloring = "cppad.symmetric";
            f_.sparse_hes(x, w, hes_subset_, hes_pattern_, coloring, hes_work_);
            hes_computed_ = true;

            size_t nnz = hes_subset_.nnz();

            Array<size_t, Shape<Dynamic>, SizeVector> row(std::move(SizeVector(hes_subset_.row())), {nnz});
            Array<size_t, Shape<Dynamic>, SizeVector> col(std::move(SizeVector(hes_subset_.col())), {nnz});
            Array<Complex, Shape<Dynamic>, ComplexVector> val(std::move(ComplexVector(hes_subset_.val())), {nnz});

            return py_sparse_matrix_type_(
                pybind11::make_tuple(std::move(val), pybind11::make_tuple(std::move(row), std::move(col))),
                pybind11::make_tuple(x.size(), x.size()));
        }

    void set_hess_pattern(PyArrayView<const int32_t,Shape<Dynamic> > row_in,
                          PyArrayView<const int32_t,Shape<Dynamic> > col_in)
        {
            size_t nnz_in = row_in.size();

            if (row_in.size() != col_in.size())
                throw std::domain_error("row/col vectors of different size");

            if (!computed_)
                throw std::runtime_error("hessian not computed");

            size_t n = hes_pattern_.nr();

            /* Modified sparsity pattern */
            CppAD::sparse_rc<SizeVector> pat(n, n, nnz_in);
            size_t nnz_out = 0;
            for (size_t k = 0; k < nnz_in; ++k) {
                size_t r = static_cast<size_t>(row_in(k));
                size_t c = static_cast<size_t>(col_in(k));
                if (hes_pattern_map_.find(std::make_pair(r, c)) != hes_pattern_map_.end())
                    pat.set(nnz_out++, r, c);
            }
            pat.resize(n, n, nnz_out);

            hes_subset_ = pat;
            hes_work_.clear();
        }

    auto get_hess_pattern()
        {
            if (!computed_)
                throw std::runtime_error("hessian not computed");

            size_t nnz = hes_pattern_.nnz();
            Array<size_t, Shape<Dynamic>, SizeVector> row(std::move(SizeVector(hes_pattern_.row())), {nnz});
            Array<size_t, Shape<Dynamic>, SizeVector> col(std::move(SizeVector(hes_pattern_.col())), {nnz});

            return pybind11::make_tuple(std::move(row), std::move(col));
        }

    bool get_hess_computed()
        {
            return hes_computed_;
        }
};

PYBIND11_MODULE(_core, m)
{
    namespace py = pybind11;

    import_array1();

    m.doc() = "usadelndsoc._core";

    py::class_<Action>(m, "Action")
        .def(py::init<double, double, Complex, Complex, Action::MaskType, Action::UType, Action::OmegaType>(), "Action",
             py::arg("L_x"), py::arg("L_y"), py::arg("alpha"), py::arg("D"), py::arg("mask"), py::arg("U"), py::arg("Omega"))
        .def("eval", &Action::eval, "Evaluate value of action", py::arg("Q"))
        .def("grad", &Action::grad, "Evaluate gradient of action", py::arg("Q"))
        .def("hess", &Action::hess, "Evaluate hessian of action", py::arg("Q"))
        .def("hess_mul", &Action::hess_mul, "Evaluate hessian(Q) * dQ", py::arg("Q"), py::arg("dQ"))
        .def("set_hess_pattern", &Action::set_hess_pattern, "Set Hessian sparsity pattern", py::arg("row"), py::arg("col"))
        .def("get_hess_pattern", &Action::get_hess_pattern, "Get original Hessian sparsity pattern")
        .def("compute", &Action::compute, "Precompute for AD", py::arg("Q"))
        .def_property("omega", &Action::get_omega, &Action::set_omega)
        .def_property_readonly("hess_computed", &Action::get_hess_computed)
        ;

    m.attr("MASK_NONE") = static_cast<int>(Mask::NONE);
    m.attr("MASK_TERMINAL") = static_cast<int>(Mask::TERMINAL);
    m.attr("MASK_VACUUM") = static_cast<int>(Mask::VACUUM);
}
